/*
 *
 * nucleo/core/carbon/cReactiveEngine.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/core/carbon/cReactiveEngine.H>
#include <nucleo/gl/window/glWindow.H>

#include <Carbon/Carbon.h>

#include <iostream>

namespace nucleo {

  static const long cReactiveEngine_event_class = 'nucE' ;
  static const long cReactiveEngine_event_kind = 'noti' ;

  OSStatus
  cReactiveEngine::cb(EventHandlerCallRef nextHandler, EventRef theEvent, void* userData) {
    ReactivePair *pair = 0 ;
    OSStatus result = GetEventParameter(theEvent, kEventParamUserData, typeVoidPtr, NULL,
								sizeof(pair), NULL, &pair) ;
    if (result==noErr && pair) {
	 if (ReactiveObject::isAlive(pair->first))
	   pair->first->react(pair->second) ;
	 else {
	   // std::cerr << "cReactiveEngine: ReactiveObject " << pair->first << " is dead..." << std::endl ;
	   cReactiveEngine *myself = (cReactiveEngine*)userData ;
	   myself->detachFrom(pair->first, pair->second) ;
	 }

	 delete pair ;
	 return noErr ;
    }

    return eventNotHandledErr ;
  }
  
  cReactiveEngine::cReactiveEngine(void) {
    EventTypeSpec ets ;
    ets.eventClass = cReactiveEngine_event_class ;
    ets.eventKind = cReactiveEngine_event_kind ;

    InstallEventHandler(GetApplicationEventTarget(),
				    NewEventHandlerUPP(cb),
				    1, &ets,
				    (void *)this,
				    NULL /* outRef */) ;
  }

  cReactiveEngine::~cReactiveEngine(void) {
  }

  void
  cReactiveEngine::notify(ReactiveObject *notified, Observable *notifier) {
    ReactivePair *pair = new ReactivePair(notified,notifier) ;
    EventRef event;
    CreateEvent(0,
			 cReactiveEngine_event_class, cReactiveEngine_event_kind,
			 0, kEventAttributeUserEvent, &event);
    SetEventParameter(event,
				  kEventParamUserData, typeVoidPtr,
				  sizeof(pair), (void *)&pair) ;

    EventQueueRef eQueue = GetMainEventQueue() ;
    EventPriority priority = kEventPriorityStandard ;
#if 0
    glWindow *window = dynamic_cast<glWindow *>(notifier) ;
    // Warning: this changes the order in which events are processed...
    if (window) priority = kEventPriorityHigh ;
#endif
    PostEventToQueue(eQueue, event, priority);
    // std::cerr << GetNumEventsInQueue(eQueue) << " events queued" << std::endl ;

    ReleaseEvent(event) ;
  }

  int
  cReactiveEngine::step(long milliseconds) {
    EventRef theEvent ;
    EventTargetRef theTarget = GetEventDispatcherTarget() ;
    EventTimeout carbon_timeout = (milliseconds==-1) ? kEventDurationForever : milliseconds*kEventDurationMillisecond ;
    if (ReceiveNextEvent(0, NULL, carbon_timeout, true, &theEvent)==noErr) {
	 SendEventToEventTarget(theEvent, theTarget) ;
	 ReleaseEvent(theEvent) ;
	 return 1 ;
    }
    return 0 ;
  }

  void
  cReactiveEngine::run(void) {
    RunApplicationEventLoop() ;
  }

  void
  cReactiveEngine::stop(void) {
    QuitApplicationEventLoop() ;
  }

}
