/*
 *
 * nucleo/plugins/ffmpeg/ffmpegImageSink.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

// FIXME: ffmpegImageSink is broken.
//  TODO: Check output_example.c in FFmpeg source tree.

#include <nucleo/config.H>

#include <nucleo/plugins/ffmpeg/ffmpegImageSink.H>

#include <nucleo/utils/FileUtils.H>
#include <nucleo/image/encoding/Conversion.H>
#include <nucleo/core/URI.H>

#include <unistd.h>

#include <stdexcept>

// ------------------------------------------------------------------------------------------------------

extern "C" {

  void *ffmpegImageSink_factory(const URI *uri)
  {
	  return new ffmpegImageSink(*uri) ;
  }

}

// ------------------------------------------------------------------------------------------------------

ffmpegImageSink::ffmpegImageSink(const URI &uri) {
  _active = false ;

  _filename = uri.isOpaque() ? uri.getOpaque() : uri.getPath() ;
  _format = "" ;

  std::string query = uri.getQuery() ;
  std::string arg;
  if (URI::getQueryArg(query, "framerate", &arg)) {
    _framerate = (arg!="") ? atoi(arg.c_str()) : 25 ;
  }
  else
    _framerate = 25 ;
  
  if (URI::getQueryArg(query, "bitrate")) {
    _bitrate = (arg!="") ? atoi(arg.c_str()) : 256000 ;
  }
  else
    _bitrate = 256000 ;

  if (URI::getQueryArg(query, "gopsize")) {
    _gopsize = (arg!="") ? atoi(arg.c_str()) : 12 ;
  }
  else
    _gopsize = 12 ; //12

  if (URI::getQueryArg(query, "qscale")) {
    _qscale = (arg!="") ? atoi(arg.c_str()) : 2 ;
  }
  else
    _qscale = 2 ;
}

// ------------------------------------------------------------------------------------------------------

bool ffmpegImageSink::init(Image *img)
{
  
	if (!first_oformat)
	{
		av_register_all() ;
		avcodec_init();
		avcodec_register_all() ;
		avienc_init();
		movenc_init();
	}

	if (_format=="")
	{
		_avf = guess_format(0,_filename.c_str(),0);
	}
	else
	{
		_avf = guess_format(_format.c_str(),0,0);
	}

	if (!_avf)
	{
		std::cerr << "ffmpegImageSink: couldn't guess format from (" <<
			_format << "," << _filename << ")" << std::endl ;
		return false ;
	}
	if (_avf->video_codec == CODEC_ID_NONE)
	{
		std::cerr << "ffmpegImageSink: '" << _avf->long_name
			  << "' has no video codec ID" << std::endl ;
		return false ;
	}

	std::cerr << "ffmpegImageSink: " << _avf->long_name << " codec "
		  << _avf->video_codec << std::endl ;

	_avfc = (AVFormatContext *)av_mallocz (sizeof(AVFormatContext));
	_avfc->oformat = _avf ;
	_avfc->priv_data = av_mallocz (_avfc->oformat->priv_data_size);

	if (!_avfc->priv_data)
	{
		std::cerr << "ffmpegImageSink: private data alloc error"
			  << std::endl ;
		return false;
	}

	/* prepare output file for format context */
        {
		char first;
		char tmp_buf[PATH_MAX+1];
            
		first = _filename[0];
		sprintf(tmp_buf, "%s", _filename.c_str());
		if (first != '/')
		{
		  char *cwd = getcwd(0,0) ;
		  if (cwd) {
			sprintf(_avfc->filename,"file://%s/%s", cwd, tmp_buf) ;
			free(cwd) ;
		  } else {
		    perror("ffmpegImageSink::init") ;
		    sprintf(_avfc->filename,"file:%s", tmp_buf) ;
		  }
		}
		else
		{
			sprintf(_avfc->filename, "file://%s", tmp_buf);
		}
        }

        register_protocol (&file_protocol);

	_st = av_new_stream (_avfc, 0);
	if (!_st)
	{
		std::cerr << "ffmpegImageSink: fail to create av steam"
			<< std::endl;
	}

	// avcodec_get_context_defaults(&(_st->codec));

	_st->codec.codec_id = _avf->video_codec;
        _st->codec.codec_type = CODEC_TYPE_VIDEO;
	_st->codec.width = img->getWidth() ;
	_st->codec.height = img->getHeight() ;
	_st->codec.bit_rate = _bitrate ;
	_st->codec.frame_rate = _framerate ;
	_st->codec.frame_rate_base = 1 ;
	_st->codec.gop_size = _gopsize ;
	//_st->codec.flags = 0 ;
	_st->codec.pix_fmt = PIX_FMT_YUV420P ;
	//_st->codec.max_b_frames=1;

	if (_qscale)
	{
		_st->codec.flags |= CODEC_FLAG_QSCALE ;
		//_st->codec.quality = _qscale ;
		_st->quality = _qscale ;
	}
	std::cerr << "ffmpegImageSink: " << _st->codec.width << " "
		  << _st->codec.height << " " << _st->codec.bit_rate << " "
		  << _st->codec.frame_rate << " " << _st->codec.gop_size
		  << std::endl ;

	// set params
	AVFormatParameters fParams, *p_fParams = &fParams;
	AVImageFormat *image_format = NULL;
        memset (p_fParams, 0, sizeof(*p_fParams));
        p_fParams->image_format = image_format;
        p_fParams->frame_rate = _st->codec.frame_rate;
        p_fParams->frame_rate_base = _st->codec.frame_rate_base;
        p_fParams->width = _st->codec.width;
        p_fParams->height = _st->codec.height;
        if (av_set_parameters (_avfc, p_fParams) < 0)
	{
            std::cerr << "ffmpegImageSink: av_set_parameters fail" << std::endl;
            return false;
        }
        

	_codec = avcodec_find_encoder(_avf->video_codec);
	//_st.codec.codec = avcodec_find_encoder(_avf->video_codec) ;
	if (!_codec)
	{
		std::cerr << "ffmpegImageSink: avcodec_find_encoder returned 0"
			  << std::endl ;
		return false ;
	}

	if (avcodec_open(&(_st->codec), /*_st.codec.codec*/_codec) < 0)
	{
		std::cerr << "ffmpegImageSink: avcodec_open failed"
			  << std::endl ;
		return false ;
	}
  
	/* open the file */
        if (url_fopen(&_avfc->pb, _avfc->filename, URL_WRONLY) < 0)
	{
		std::cerr << "ffmpegImageSink: Could not open " <<
			_avfc->filename << std::endl ;
		return false;
        }
        
        if (av_write_header (_avfc) < 0)
	{
		dump_format (_avfc, 0, _avfc->filename, 1);
		std::cerr << "ffmpegImageSink: Could not write header" <<
			std::endl;
		return false;
        }

	_frame = avcodec_alloc_frame();
	_frame->type = FF_BUFFER_TYPE_SHARED;

	_frame->linesize[0] = _st->codec.width ;
	_frame->linesize[1] = _st->codec.width / 2 ;
	_frame->linesize[2] = _st->codec.width / 2 ;
	_frame->quality = (int)_st->quality;

	int image_size = avpicture_get_size(
		PIX_FMT_YUV420P, _st->codec.width, _st->codec.height);
	_outbuffer = (unsigned char *)malloc (image_size);

	return true ;
}

bool
ffmpegImageSink::start(void) {
  _active = true ;
  _chrono.start() ;
  return true ;
}

bool
ffmpegImageSink::handle(Image *img)
{
	Image copy(*img);
	if (!convertImage(&copy, Image::YpCbCr420))
	{
		return false ;
	}

	//int w = copy.getWidth(), h=copy.getHeight() ;

	if (!_chrono.count()) _active = init(&copy) ;
	if (!_active) return false ;

	// --------------------------------------------

	_frame->data[0] = copy.getData() ;
	int size = _st->codec.width * _st->codec.height;
	_frame->data[1] = _frame->data[0] + size ;
	_frame->data[2] = _frame->data[1] + size / 4 ;
	_frame->data[3] = NULL;

	int nb = avcodec_encode_video(
		&_st->codec, _outbuffer, sizeof(_outbuffer), _frame) ;

	 av_write_frame (_avfc, _st->index, _outbuffer, nb);

	_chrono.tick() ;

	return true ;
}

bool
ffmpegImageSink::stop(void) {
	bool result = false ;

	if (_active)
	{
		if (_st)
		{
			avcodec_close(&(_st->codec));
			_st = 0;
		}
		if (_avfc)
		{
			av_write_trailer (_avfc);
			for(int j=0;j<_avfc->nb_streams;j++)
				av_free (_avfc->streams[j]);
			/* close file */
			url_fclose (&_avfc->pb);
			/* free format context */
			av_free(_avfc);
			_avfc = NULL;
		}
		_chrono.stop();
		std::cerr << "ffmpegImageSink: "<< _chrono.average() << " fps"
			  << std::endl;  
		_active = false ;
		result = true ;
	}

	return result ;
}
