/*
 *
 * nucleo/plugins/howl/howlServiceAnnouncer.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/plugins/howl/howlServiceAnnouncer.H>

#include <iostream>

extern "C" {
  nucleo::howlServiceAnnouncer*
  howlServiceAnnouncer_factory(const char *n, const char *t, int p,
						 nucleo::TextRecords *textRecords,
						 bool ar,
						 const char *d, const char *h) {
    return new nucleo::howlServiceAnnouncer(n,t,p,textRecords,ar,d,h) ;
  }
}

namespace nucleo {

  sw_result
  howlServiceAnnouncer::saServiceReply(sw_discovery discovery,
							    sw_discovery_publish_status status,
							    sw_discovery_oid id,
							    sw_opaque extra) {
    howlServiceAnnouncer *obj = (howlServiceAnnouncer *)extra ;
    obj->status = status ;
#if 0
    const char *snames[] = {"SW_DISCOVERY_PUBLISH_STARTED",
					   "SW_DISCOVERY_PUBLISH_STOPPED",
					   "SW_DISCOVERY_PUBLISH_NAME_COLLISION",
					   "SW_DISCOVERY_PUBLISH_INVALID"} ;
    std::cerr << "status = " << status << " (" << snames[status] << ")" << std::endl ;
#endif
    if (obj->autorename && status==SW_DISCOVERY_PUBLISH_NAME_COLLISION)
	 obj->changeName() ;

    return SW_OKAY ;
  }
 
  howlServiceAnnouncer::howlServiceAnnouncer(const char *n, const char *t, int p,
									TextRecords *textRecords,
									bool ar,
									const char *d, const char *h) {
    autorename = ar ;

    name = strdup(n) ;
    type = strdup(t) ;
    port = p ;
    domain = d ? strdup(d) : 0 ;
    host = h ? strdup(h) : 0 ;

    howlKeeper::attach() ;

    if (sw_text_record_init(&text_record) != SW_OKAY)
	 throw std::runtime_error("sw_text_record_init() failed");
    if (textRecords) {
	 for (TextRecords::iterator pair=textRecords->begin();
		 pair!=textRecords->end();
		 ++pair) {
	   std::string key = (*pair).first ;
	   std::string value = (*pair).second ;
	   if (sw_text_record_add_key_and_string_value(text_record,
										  key.c_str(), value.c_str())!=SW_OKAY)
		std::cerr << "Unable to add service text record: " << key << "=" << value << std::endl ;
	 }
    }
   
    announce() ;
  }

  void
  howlServiceAnnouncer::announce(void) {
    status = SW_DISCOVERY_PUBLISH_INVALID ;
    sw_discovery_publish(howlKeeper::getSession(),
					0,
					name, type, domain, host, port,
					sw_text_record_bytes(text_record),
					sw_text_record_len(text_record),
					saServiceReply, this,
					&id) ;
  }

  howlServiceAnnouncer::state
  howlServiceAnnouncer::getState(void) {
    switch (status) {
    case SW_DISCOVERY_PUBLISH_NAME_COLLISION: return NAME_COLLISION ;
    case SW_DISCOVERY_PUBLISH_STARTED: return ANNOUNCED ;
    default: return UNKNOWN ;
    }
  }

  bool
  howlServiceAnnouncer::changeName(const char *newname) {
    if (status!=SW_DISCOVERY_PUBLISH_NAME_COLLISION) return false ;

    if (newname) {
	 free(name) ;
	 name = strdup(newname) ;
    } else {
	 std::string newname = name ;
	 if (name[strlen(name)-1]!='+') newname = newname + " " ;
	 newname = newname + "+" ;
	 free(name) ;
	 name = strdup(newname.c_str()) ;
    }

    announce() ;
    return true ;
  }

  std::string
  howlServiceAnnouncer::getName(void) {
    return name ;
  }

  howlServiceAnnouncer::~howlServiceAnnouncer(void) {
    sw_text_record_fina(text_record) ;
    sw_discovery_cancel(howlKeeper::getSession(), id) ;
    free(name) ;
    free(type) ;
    if (domain) free(domain) ;
    if (host) free(host) ;
    howlKeeper::detach() ;
  }

}
