/*
 *
 * nucleo/plugins/howl/howlServiceFinder.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/plugins/howl/howlServiceFinder.H>

#include <discovery/text_record.h>

#include <iostream>

extern "C" {
  nucleo::howlServiceFinder*
  howlServiceFinder_factory(const char *type, nucleo::ServiceFinder::EventQueue *queue, const char *domain) {
    return new nucleo::howlServiceFinder(type,queue,domain) ;
  }
}

namespace nucleo {

  sw_result
  howlServiceFinder::sfResolver(sw_discovery discovery,
						  sw_discovery_oid id,
						  sw_uint32 interface_index,
						  sw_const_string name,
						  sw_const_string type,
						  sw_const_string domain,
						  sw_ipv4_address address,
						  sw_port port,
						  sw_octets text_record,
						  sw_ulong text_record_len,
						  sw_opaque extra) {
    sw_discovery_cancel(discovery, id) ;

    howlServiceFinder *obj = (howlServiceFinder*)extra ;

    obj->rids.remove(id) ;

    Event *e = new Event ;
    e->event = Event::FOUND ;
    e->service.type = type ;
    e->service.name = name ;
    e->service.domain = domain ;
    char name_buf[16] ; sw_ipv4_address_name(address, name_buf, 16) ;
    e->service.ipaddr = name_buf ;
    e->service.port = port ;    

    sw_text_record_iterator it ;
    if (sw_text_record_iterator_init(&it, text_record, text_record_len) == SW_OKAY) {
	 char key[255] ;
	 sw_uint8 val[255] ;
	 sw_uint32 val_len ;
	 while (sw_text_record_iterator_next(it, key, val, &val_len) == SW_OKAY) {
	   std::string value((const char *)val,0,val_len) ;
	   e->service.textRecords[key] = value ;
	 }
    }
    sw_text_record_iterator_fina(it) ;

    obj->queue->push(e) ;
    obj->notifyObservers() ;

    return SW_OKAY ;
  }

  sw_result
  howlServiceFinder::sfBrowseReply(sw_discovery discovery,
			   sw_discovery_oid id,
			   sw_discovery_browse_status status,
			   sw_uint32 interface_index,
			   sw_const_string name,
			   sw_const_string type,
			   sw_const_string domain,
			   sw_opaque extra) {
    howlServiceFinder *obj = (howlServiceFinder*)extra ;

    switch (status) {
    case SW_DISCOVERY_BROWSE_ADD_SERVICE: {
	 sw_discovery_resolve_id rid;
	 if (sw_discovery_resolve(discovery,
						 0,
						 name, type, domain,
						 sfResolver, extra, &rid) != SW_OKAY)
	   std::cerr << "howlServiceFinder: sw_discovery_resolve failed" << std::endl ;
	 else
	   obj->rids.push_back(rid) ;
    } break;
    case SW_DISCOVERY_BROWSE_REMOVE_SERVICE: {
	 Event *e = new Event ;
	 e->event = Event::LOST ;
	 e->service.type = type ;
	 e->service.name = name ;
	 e->service.domain = domain ;
	 obj->queue->push(e) ;
	 obj->notifyObservers() ;
    } break;
    default:
	 break;
    }

    return SW_OKAY;
  }

  howlServiceFinder::howlServiceFinder(const char *type, EventQueue *q, const char *domain)
    : ServiceFinder(q) {
    howlKeeper::attach() ;
    if (sw_discovery_browse(howlKeeper::getSession(),
					   0,
					   type, domain,
					   sfBrowseReply, this, &id) != SW_OKAY)
	 throw std::runtime_error("howlServiceFinder: sw_discovery_browse failed") ;
  }

  howlServiceFinder::~howlServiceFinder(void) {
    sw_discovery session = howlKeeper::getSession() ;
    sw_discovery_cancel(session, id) ;
    for (std::list<sw_discovery_resolve_id>::iterator i=rids.begin();
	    i!=rids.end();
	    ++i)
	 sw_discovery_cancel(session, *i) ;
    howlKeeper::detach() ;
  }

}
